#include "stdafx.h"
#pragma hdrstop

HWND		g_hWnd;
HINSTANCE	g_hInst;

const char * _className = "Orange_Class_Wnd";
const char * _windowName = "OpenGL 3.0";

const int _WND_WIDTH	=	800;
const int _WND_HEIGHT	=	600;

template<int what, int to>
int change_char( int c ) { if ( what == c ) return to; return c; }

LRESULT CALLBACK WndProc( HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam );
bool CreateMainWindow( int w, int h );
void InitVertexData( void );
void Draw( void );


GLint		_modelViewProjection;
GLint		_vertPosition;
GLint		_vertColor;
GLuint		_vbo;

uint32 _prevTime;
float  _curAngle = 0;
mat4   proj, mv, mvp;


int main( int argc, char* argv[] )
{
	g_hInst = GetModuleHandle( NULL );

	Log_Normal( "Creating main window..." );
	if ( !CreateMainWindow( _WND_WIDTH, _WND_HEIGHT ) )
	{
		Log_Error( "Couldn't create main window" );
		return 1;
	}

	HGLRC _glRC;
	Log_Normal( "Trying to initialize OpenGL 3.0" );
	if ( !OpenGL_3::Initialize( g_hWnd, _glRC ) )
	{
		Log_Error( "Couldn't init OpenGL 3.0... Shutdown..." );
		return 1;
	}
	OpenGL_3::AssertGLError();

	std::string oglversion = OpenGL_3::GetVersion();
	OpenGL_3::AssertGLError();
	std::string ogl_vendor = OpenGL_3::GetVendor();
	OpenGL_3::AssertGLError();
	std::string ogl_renderer = OpenGL_3::GetRenderer();
	OpenGL_3::AssertGLError();
	std::string extensions = OpenGL_3::GetExtensions();
	OpenGL_3::AssertGLError();
	std::string wgl_extens = OpenGL_3::GetWGLExtensions();
	OpenGL_3::AssertGLError();
	std::transform( extensions.begin(), extensions.end(), extensions.begin(), change_char<' ','\n'> );
	std::transform( wgl_extens.begin(), wgl_extens.end(), wgl_extens.begin(), change_char<' ','\n'> );
	std::cout << "OpenGL Version  :  " << oglversion << std::endl;
	std::cout << "OpenGL Vendor   :  " << ogl_vendor << std::endl;
	std::cout << "OpenGL Renderer :  " << ogl_renderer << std::endl;
	std::cout << "OpenGL Extentions:" << std::endl << extensions << std::endl;
	std::cout << "OpenGL WGL-Extentions:" << std::endl << wgl_extens << std::endl;

	Log_Normal( "Initializing VBO extensions..." );
	if ( !glext::InitVBO() )
	{
		Log_Error( "Couldn't initialize VBO" );
		OpenGL_3::Shutdown();
	}
	Log_Normal( "Initializing VBO extensions OK" );
	glext::InitVertexAttribs();
	Log_Normal( "Initializing GLSL extensions..." );
	if ( !glext::InitGLSLNeededExts() )
	{
		Log_Error( "Couldn't initialize GLSL" );
		OpenGL_3::Shutdown();
	}
	Log_Normal( "Initializing GLSL extensions OK" );
	std::cout << "OpenGL GLSL Version :  " << OpenGL_3::GetGLSLVersion() << std::endl;

	const std::string vs =	"#version 130\n"\
							"\tprecision highp float;\n"\
							"\tuniform mat4x4 modelViewProjection;\n"\
							"\tin vec4 vertPosition;\n"\
							"\tin vec4 vertColor;\n"\
							"\tout vec4 color;\n"\
							"\tvoid main() {\n"\
							"\t\tcolor = vertColor;\n"\
							"\t\tgl_Position = modelViewProjection * vertPosition;\n"\
							"\t}\n";
	const std::string fs =	"#version 130\n"\
							"\tprecision highp float;\n"\
							"\tin vec4 color;\n"\
							"\tout vec4 fragColor;\n"\
							"\tvoid main() {\n"\
							"\tfragColor = color;\n"\
							"\t}\n";

	Log_Normal( "Try to compile such shader:\nVertex Program:\n\t" + vs + "Fragment Program:\n\t" + fs );
	GLhandleARB program = OpenGL_3::CreateGLSLProgram( vs, fs );
	if ( NULL == program )
	{
		Log_Error( "Compilation failed... Exiting..." );
		exit( 1 );
	}
	Log_Normal( "Compilation ok!" );

	glext::glUseProgramObjectARB( program );
	_modelViewProjection = glext::glGetUniformLocationARB( program, "modelViewProjection" );
	_vertPosition = glext::glGetAttribLocationARB( program, "vertPosition" );
	_vertColor = glext::glGetAttribLocationARB( program, "vertColor" );

	glViewport( 0, 0, _WND_WIDTH, _WND_HEIGHT );
	glScissor( 0, 0, _WND_WIDTH, _WND_HEIGHT );
	glClearColor( 0, 0, 0, 1 );

	proj = MakeOrtho2D( 0, (float)_WND_WIDTH, (float)_WND_HEIGHT, 0, -1, 1 );
	mvp = mv * proj;

	glext::glUniformMatrix4fvARB( _modelViewProjection, 1, 0, mvp );

	InitVertexData();

	OpenGL_3::AssertGLError();

	_prevTime = GetTickCount();
	
	MSG msg;
	while ( true )
	{
		if ( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) )
		{
			if ( !GetMessage( &msg, NULL, 0, 0 ) || WM_QUIT == msg.message )
				break;
			else
			{
				TranslateMessage( &msg );
				DispatchMessage ( &msg );
			}
		}

		Draw();
		Sleep(1);
	}

	OpenGL_3::Shutdown();
	return 0;
}

LRESULT CALLBACK WndProc( HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam )
{
	switch ( wMsg )
	{
		case WM_KEYDOWN:
		{
			if ( VK_ESCAPE == wParam )
				PostQuitMessage(0);
		} return 0;

		case WM_DESTROY:
		case WM_CLOSE:
		{
			PostQuitMessage(0);
		} return 0;

		default:
			return DefWindowProc( hWnd, wMsg, wParam, lParam );
	}
}

bool CreateMainWindow( int w, int h )
{
	WNDCLASSEX wc;
	RtlZeroMemory( &wc, sizeof( wc ) );
	wc.cbSize        = sizeof( wc );
	wc.style         = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc   = (WNDPROC)WndProc;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = GetModuleHandle( NULL );
	wc.hIcon         = LoadIcon( NULL, IDI_WINLOGO );
	wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground = NULL;
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = _className;

	if ( !RegisterClassEx( &wc ) )
	{
		Log_Error( "Couldn't register wnd class" );
		return false;
	}
	uint32 styleEx = 0;
	uint32 style   = (WS_VISIBLE | WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN) ^ WS_MAXIMIZEBOX;
	RECT rc;
	SetRect( &rc, 0, 0, w, h );
	AdjustWindowRectEx( &rc, style, FALSE, styleEx );
	if ( !( g_hWnd = CreateWindowEx( styleEx, _className, _windowName, style, 0, 0,
									 rc.right - rc.left, rc.bottom - rc.top, 0, 0, g_hInst, 0 ) ) )
	{
		Log_Error( "Couldn't create main window" );
		UnregisterClass( _className, g_hInst );
		return false;
	}

	return true;
}



struct DrawVertex_s
{
	vec3	pos;
	uint8	r, g, b, a;
};

void InitVertexData( void )
{
	DrawVertex_s vertexes[] =
	{
		{ vec3(  50, 500, 0 ),	255, 0, 0, 255 },
		{ vec3( 750, 500, 0 ),	0, 255, 0, 255 },
		{ vec3( 400,  50, 0 ),	0, 0, 255, 255 }
	};

	glext::glGenBuffersARB( 1, &_vbo );
	glext::glBindBufferARB( GL_ARRAY_BUFFER, _vbo );
		glext::glBufferDataARB( GL_ARRAY_BUFFER, sizeof(vertexes), vertexes, GL_STATIC_DRAW );
	glext::glBindBufferARB( GL_ARRAY_BUFFER, 0 );
}

#define vbo_offsetof(s,m)	((GLvoid*)(offsetof(s,m)))

void Draw( void )
{
	glClear( GL_COLOR_BUFFER_BIT );

	glext::glBindBufferARB( GL_ARRAY_BUFFER, _vbo );

	glext::glEnableVertexAttribArrayARB( _vertColor );
	glext::glEnableVertexAttribArrayARB( _vertPosition );

	glext::glVertexAttribPointerARB( _vertColor, 4, GL_UNSIGNED_BYTE, GL_TRUE, sizeof(DrawVertex_s), vbo_offsetof(DrawVertex_s, r) );
	glext::glVertexAttribPointerARB( _vertPosition, 3, GL_FLOAT, GL_FALSE, sizeof(DrawVertex_s), NULL );


	uint32 deltaTime = GetTickCount() - _prevTime;
	_prevTime = GetTickCount();
	_curAngle += float(deltaTime)*0.05f;
	if ( _curAngle > 360.0f )
		_curAngle -= 360.0f;

	mv = RotateMatrixZ( _curAngle );
	mvp = mv * proj;
	glext::glUniformMatrix4fvARB( _modelViewProjection, 1, 0, mvp );

	glDrawArrays( GL_TRIANGLES, 0, 3 );

	SwapBuffers( wglGetCurrentDC() );
}
